/*
 *  R : A Computer Langage for Statistical Data Analysis
 *  Copyright (C) 1995  Robert Gentleman and Ross Ihaka
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "Defn.h"
#include <string.h>
#include <stdlib.h>


#ifdef DLSupport

	/* This code adds dynamic loading of code to R.  It */
	/* replaces the original dotcode.c and contains code */
	/* from that file. This code should work on a system */
	/* that implements the Solaris dlopen interface */
	/* standard. Shared libraries are loaded by dlopen */
	/* and symbols are looked for by dlsym. Unloading of */
	/* code is not implemented. This code was developed */
	/* on ELF-Linux. */
	/* Modifications Copyright (C) 1996 Heiner Schwarte */
	/* Minor cleanup and bug fix (C) 1996 Ross Ihaka */


#include <dlfcn.h>

	/* The dlopen function returns a handle after successfully */
	/* loading a library. These handles are collected in a list. */

struct voidptrlist{
	void *ptr;
	struct voidptrlist *next;
};

static struct voidptrlist *list=NULL;

static void add_ptr(void *p)
{
	struct voidptrlist *tmp;
	if (p==NULL) 
		return;
	tmp=(struct voidptrlist *)malloc(sizeof(struct voidptrlist));
	tmp->ptr=p;
	tmp->next=list;
	list=tmp;
}


	/* findDynProc checks whether one of the libraries */
	/* that have been loaded contains the symbol name and */
	/* returns a pointer to that symbol upon success. */

int (*findDynProc(char *name))()
{
	struct voidptrlist *tmp;
	void *fcnptr;
	char buf[64];

	sprintf(buf, "_%s", name);
	tmp = list;
	while(tmp != NULL){
		if((fcnptr = dlsym(tmp->ptr, buf)) != NULL)
			return fcnptr;
		tmp = tmp->next;
	}
	return NULL;
}


	/* do_dynload implements the R-Interface for the */
	/* loading of libraries */

#ifndef RTLD_LAZY
#define RTLD_LAZY 1
#endif

SEXP do_dynload(SEXP call, SEXP op, SEXP args, SEXP env)
{
	void *handle;
	char *error;
	checkArity(op,args);
	if (!isString(CAR(args)))
		errorcall(call, "character argument expected\n");
	handle = dlopen(CHAR(STRING(CAR(args))[0]), RTLD_LAZY);
	if(!handle) {
		error = dlerror();
		errorcall(call, error);
		free(error);
	}
	add_ptr(handle);
	R_Visible = 0;
	return R_NilValue;
}

#else

SEXP do_dynload(SEXP call, SEXP op, SEXP args, SEXP env)
{
	error("no dyn.load support in this R version\n");
}

#endif




#define NIL -1

typedef int (*FUNC) ();

typedef struct {
	char *name;
	FUNC func;
} CFunTabEntry;

#ifdef no_f77_underscore
# define F77_SUBROUTINE(x) int x ## ();
#else
# define F77_SUBROUTINE(x) int x ## _();
#endif
#define C_FUNCTION(x) int x ## ();
#include "dotcode.h"
#undef F77_SUBROUTINE
#undef C_FUNCTION

#ifdef no_f77_underscore
#define F77_SUBROUTINE(x) {#x, x},
#else
#define F77_SUBROUTINE(x) {#x ## "_", x ## _ },
#endif
#define C_FUNCTION(x) {#x, x },
static CFunTabEntry CFunTab[] =
{
#include "dotcode.h"
	{NULL, NULL}
};
#undef F77_SUBROUTINE
#undef C_FUNCTION


	/* HashTable stores name - pointer pairs. Open addressing */
	/* with linear probing is used. Sometimes the hashtable */
	/* will be expanded and reorganized. The implementation */
	/* is entirely elementary.  Possible sizes of the */
	/* table are 2^p+1 where p is a positive integer. */

static CFunTabEntry *HashTable;
static int HASHSIZE;
static int NumberElem;


static int HashCode(char *symbol)
{
	unsigned int  code=0;
	char *p = symbol;
  
	while (*p)
		code = 8 * code + *p++;
	return code % HASHSIZE;
}


static void HashInstall(char *name, FUNC func)
{
	int key;
	NumberElem++;
	key = HashCode(name);
	while (HashTable[key].name != NULL)
		key = (key + 1) % HASHSIZE;
	HashTable[key].name = (char*)malloc(strlen(name)+1);
	strcpy(HashTable[key].name,name);
	HashTable[key].func = func;
}

static void HashExpand()
{
	int oldsize;
	int i;
	CFunTabEntry *OldTable;
	oldsize=HASHSIZE;
	OldTable=HashTable;
	HASHSIZE=2*HASHSIZE-1;
	NumberElem=0;
	HashTable = (CFunTabEntry *) malloc(HASHSIZE * sizeof(CFunTabEntry));
	for (i = 0; i < HASHSIZE; i++)
		HashTable[i].name = NULL;
	for(i=0;i<oldsize;i++) {
		if(OldTable[i].name!=NULL)
			HashInstall(OldTable[i].name,OldTable[i].func);
	}	
	for(i=0;i<oldsize;i++) 
		free(OldTable[i].name);
	free(OldTable);
}

static FUNC HashLookup(char *symbol)
{
	int key;
	key = HashCode(symbol);
	while (HashTable[key].name != NULL) {
		if (strcmp(symbol, HashTable[key].name) == 0)
			return HashTable[key].func;
		else
			key = (key + 1) % HASHSIZE;
	}
	return NULL;
}


/* Initialization */

void InitFunctionHashing()
{
	int n;
	int i, size = 3;
	n = sizeof(CFunTab)/sizeof(CFunTabEntry);
	while(size < n) 
		size = 2*size-1;
	HASHSIZE = size;	
	NumberElem = 0;
	HashTable = (CFunTabEntry*) malloc(HASHSIZE * sizeof(CFunTabEntry));
	for (i = 0; i < HASHSIZE; i++)
		HashTable[i].name = NULL;
	for (i = 0; CFunTab[i].name; i++)
		HashInstall(CFunTab[i].name, CFunTab[i].func);
	HashExpand();
}


/* Call a C/Fortran function from R */
SEXP do_dotCode(SEXP call, SEXP op, SEXP args, SEXP env)
{
	double *cargs[20];
	int which;
	int nargs = 0;
	FUNC fun;
	SEXP topofargs;
	char buf[128], *p, *q, *vmax;
	
	vmax = vmaxget();
	which = PRIMVAL(op);
	PROTECT(op = CAR(args));
	PROTECT(args = CDR(args));
	topofargs = args;
	
	/* duplicate the arg list */
	/* could use reference counting here */
	
	for (nargs = 0; args != R_NilValue; nargs++) {
		if (!isVector(CAR(args)))
			errorcall(call, "only vectors allowed as C/Fortran args\n");
		CAR(args) = duplicate(CAR(args));
		args = CDR(args);
	}
	
	/* set up the calling args to point to the right */
	/* part of the vector heap, NO allocing or duplication */
	/* after this point */
	
	args = topofargs;
	for (nargs = 0; args != R_NilValue; nargs++, args = CDR(args)) {
		cargs[nargs] = REAL(CAR(args));
	}
	
	
	if (!isString(op))
		errorcall(call, "function name must be a string\n");
	
	/* make up load symbol & look it up */
	
	p = CHAR(STRING(op)[0]);
	q = buf;
	while ((*q = *p) != '\0') {
		p++;
		q++;
	}
#ifdef no_f77_underscore
#else
	if (which)
		*q++ = '_';
	*q = '\0';
#endif 
	if (!(fun = HashLookup(buf))){
#ifdef DLSupport
		if(!(fun = findDynProc(buf))) {
			errorcall(call, "C/Fortran function not in load table\n");
		} 
		else {
			if((1.0*NumberElem)/HASHSIZE > 0.5)
				HashExpand();
			HashInstall(buf,fun);
		}
#else
		errorcall(call, "C/Fortran function not in load table\n");
#endif
	}
	
	switch (nargs) {
	case 0:
		/* Silicon graphics C barfs if there is */
		/* no argument to fun */
		fun(0);
		break;
	case 1:
		fun(cargs[0]);
		break;
	case 2:
		fun(cargs[0], cargs[1]);
		break;
	case 3:
		fun(cargs[0], cargs[1], cargs[2]);
		break;
	case 4:
		fun(cargs[0], cargs[1], cargs[2], cargs[3]);
		break;
	case 5:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4]);
		break;
	case 6:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5]);
		break;
	case 7:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6]);
		break;
	case 8:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7]);
		break;
	case 9:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7], cargs[8]);
		break;
	case 10:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7], cargs[8], cargs[9]);
		break;
	case 11:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7], cargs[8], cargs[9], cargs[10]);
		break;
	case 12:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7], cargs[8], cargs[9], cargs[10], cargs[11]);
		break;
	case 13:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7], cargs[8], cargs[9], cargs[10], cargs[11], cargs[12]);
		break;
	case 14:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7], cargs[8], cargs[9], cargs[10], cargs[11], cargs[12], cargs[13]);
		break;
	case 15:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7], cargs[8], cargs[9], cargs[10], cargs[11], cargs[12], cargs[13], cargs[14]);
		break;
	case 16:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7], cargs[8], cargs[9], cargs[10], cargs[11], cargs[12], cargs[13], cargs[14], cargs[15]);
		break;
	case 17:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7], cargs[8], cargs[9], cargs[10], cargs[11], cargs[12], cargs[13], cargs[14], cargs[15], cargs[16]);
		break;
	case 18:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7], cargs[8], cargs[9], cargs[10], cargs[11], cargs[12], cargs[13], cargs[14], cargs[15], cargs[16], cargs[17]);
		break;
	case 19:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7], cargs[8], cargs[9], cargs[10], cargs[11], cargs[12], cargs[13], cargs[14], cargs[15], cargs[16], cargs[17], cargs[18]);
		break;
	case 20:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7], cargs[8], cargs[9], cargs[10], cargs[11], cargs[12], cargs[13], cargs[14], cargs[15], cargs[16], cargs[17], cargs[18], cargs[19]);
		break;
	case 21:
		fun(cargs[0], cargs[1], cargs[2], cargs[3], cargs[4], cargs[5], cargs[6], cargs[7], cargs[8], cargs[9], cargs[10], cargs[11], cargs[12], cargs[13], cargs[14], cargs[15], cargs[16], cargs[17], cargs[18], cargs[19], cargs[20]);
		break;
	default:
		errorcall(call, "too many arguments, sorry\n");
	}
	UNPROTECT(2);
	vmaxset(vmax);
	return (topofargs);
}
